/* Generate SystemC code from a meta-model AST to a file.

Metropolis: Design Environment for Heterogeneus Systems.

Copyright (c) 1998-2005 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in all
copies of this software and that appropriate acknowledgments are made
to the research of the Metropolis group.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.


METROPOLIS_COPYRIGHT_VERSION_1
COPYRIGHTENDKEY


*/


package metropolis.metamodel.backends.systemc;

import metropolis.metamodel.MetaModelStaticSemanticConstants;
import metropolis.metamodel.TNLManip;
import metropolis.metamodel.TreeNode;
import metropolis.metamodel.nodetypes.ConstraintBlockNode;
import metropolis.metamodel.nodetypes.ELOCConstraintCallNode;
import metropolis.metamodel.nodetypes.ELOCConstraintDeclNode;
import metropolis.metamodel.nodetypes.ELOCConstraintNode;
import metropolis.metamodel.nodetypes.EqualVarsNode;
import metropolis.metamodel.nodetypes.LOCConstraintCallNode;
import metropolis.metamodel.nodetypes.LOCConstraintDeclNode;
import metropolis.metamodel.nodetypes.LOCConstraintNode;
import metropolis.metamodel.nodetypes.LTLConstraintCallNode;
import metropolis.metamodel.nodetypes.LTLConstraintDeclNode;
import metropolis.metamodel.nodetypes.LTLConstraintNode;
import metropolis.metamodel.nodetypes.LTLSynchNode;
import metropolis.metamodel.nodetypes.MaxDeltaNode;
import metropolis.metamodel.nodetypes.MaxRateNode;
import metropolis.metamodel.nodetypes.MinDeltaNode;
import metropolis.metamodel.nodetypes.MinRateNode;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;


//////////////////////////////////////////////////////////////////////////
//// SynchCodegenVisitor
/**
 * Generate SystemC code from a meta-model AST to a file.
 * <p>
 * Portions of this code were derived from sources developed under the
 * auspices of the Ptolemy II project.
 *
 * @author   Guang Yang
 * @version  $Id: SynchCodegenVisitor.java,v 1.13.4.1 2005/02/14 18:50:11 cxh Exp $
 */
public class SynchCodegenVisitor extends SystemCCodegenVisitor
    implements MetaModelStaticSemanticConstants{

    ///////////////////////////////////////////////////////////////////
    ////                           constructors                    ////

    /**
     * Set the traversal method to custom so that indent level can be
     * properly set.
     *
     * @param indtLevel current indent level for generated code.
     */
    public SynchCodegenVisitor(int indtLevel) {
        super();
        indentLevel =indtLevel;
    }

    /**
     * Do the same as the one-argument constructor, and allow turning on
     * the output of "<code>#line &lt;line_number&gt; &lt;file_name&gt</code>"
     * directives in the cpp code for debugging metamodel code.
     *
     * @param indtLevel current indent level for generated code.
     * @param emittingLineDirectives <code>true</code> to turn on output of
     *        <code>#line</code> directives.
     */
    //public SynchCodegenVisitor(int indtLevel, boolean emittingLineDirectives) {
    //super(emittingLineDirectives);
    //indentLevel = indtLevel;
    //}

    ///////////////////////////////////////////////////////////////////
    ////                         public variables                  ////

    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////
    private LinkedList registerSynch = new LinkedList();
    private LinkedList resolveSynch  = new LinkedList();

    ///////////////////////////////////////////////////////////////////
    ////                          public methods                   ////
    /**
     * Returns the code represented by an <code>ActionConstraintNode</code>.
     *
     * @param node      the node that is being visited
     * @param args      a list of arguments to this visit method
     * @return          a list of strings that represents
     *                  <code>constraint { STATEMENTS }</code>
     */

    public Object visitConstraintBlockNode (ConstraintBlockNode node,
            LinkedList args) {
        LinkedList retList = new LinkedList();

        //retList.add("\n");
        //retList.add(indent(indentLevel));
        //retList.add("void synch(int mode) {\n");
        retList.add(TNLManip.traverseList(this, args, node.getStmts()));
        increaseIndent();
        retList.add(indent(indentLevel));
        retList.add("if (mode == 0) {\n");
        retList.add(registerSynch);
        retList.add(indent(indentLevel));
        retList.add("} else {\n");
        retList.add(resolveSynch);
        retList.add(indent(indentLevel));
        retList.add("}\n");
        decreaseIndent();
        //retList.add(indent(indentLevel));
        //retList.add("}\n");

        return retList;
    }

    /**
     * Returns the code represented by a <code>LTLSynchNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method
     * @return                 a list of strings that represents
     *                         <code> ltl synch(events, equalVars) </code>
     */
    public Object visitLTLSynchNode(LTLSynchNode node, LinkedList args) {
        LinkedList retList = new LinkedList();
        List events = node.getEvents();
        List equVars = node.getEqualVars();
        Iterator iter = events.iterator();
        retList.addLast(indent(indentLevel));
        retList.addLast("if (\n");
        registerSynch.addLast(indent(indentLevel+2));
        registerSynch.addLast("_sb.registerSynchEvents(");
        increaseIndent();
        while (iter.hasNext()) {
            Object event = ((TreeNode) iter.next()).accept(this, args);
            retList.addLast(indent(indentLevel));
            retList.addLast(event);
            retList.addLast("->getProcess()->pc->containsPendingEvent(");
            retList.addLast(event);
            retList.addLast(")");
            registerSynch.addLast(event);
            if (iter.hasNext()) {
                retList.addLast(" && \n");
                registerSynch.addLast(", ");
            }
        }
        retList.addLast("\n");
        decreaseIndent();
        retList.addLast(indent(indentLevel));
        retList.addLast(") {\n");
        registerSynch.addLast(");\n");

        increaseIndent();
        if (equVars.size()>0) {
            iter = equVars.iterator();
            while (iter.hasNext()) {
                EqualVarsNode evNode = (EqualVarsNode) iter.next();
                //Fix me: finish this part
                //which is to compare the condition for mapped behavior
            }
        }
        decreaseIndent();

        retList.addLast(indent(indentLevel));
        retList.addLast("} else {\n");
        increaseIndent();
        iter = events.iterator();
        while (iter.hasNext()) {
            Object event = ((TreeNode) iter.next()).accept(this, args);
            retList.addLast(indent(indentLevel));
            retList.addLast("if ( ");
            retList.addLast(event);
            retList.addLast("->getProcess()->pc->containsPendingEvent(");
            retList.addLast(event);
            retList.addLast(") )\n");
            increaseIndent();
            retList.addLast(indent(indentLevel));
            retList.addLast(event);
            retList.addLast("->getProcess()->pc->SchedState = SchedStateVal::DONTRUN;\n");
            decreaseIndent();
        }
        decreaseIndent();
        retList.addLast(indent(indentLevel));
        retList.addLast("}\n\n");

        resolveSynch.addLast(retList);
        return null;
    }

    /**
     * Do not generate code represented by an <code>LTLConstraintNode</code>.
     *
     * @param node      the node that is being visited
     * @param args      a list of arguments to this visit method
     * @return          a list of strings that represents
     *                  <code>ltl(FORMULA);</code>
     */
    public Object visitLTLConstraintNode ( LTLConstraintNode node,
            LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by a <code>LOCConstraintNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method
     * @return                 a list of strings that represents
     *                         <code> loc (formula); </code>
     *                         <code> loc BuiltInLOCFormula; </code>
     */
    public Object visitLOCConstraintNode(LOCConstraintNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by a <code>ELOCConstraintNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method
     * @return                 a list of strings that represents
     *                         <code> eloc (formula); </code>
     */
    public Object visitELOCConstraintNode(ELOCConstraintNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by a <code>ELOCConstraintDeclNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method

     * @return                 a list of strings that represents
     *                         <code> eloc name ( params ) ( formula ); </code>
     */
    public Object visitELOCConstraintDeclNode(ELOCConstraintDeclNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by a <code>LOCConstraintDeclNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method
     * @return                 a list of strings that represents
     *                         <code> loc name ( params ) ( formula ); </code>
     */
    public Object visitLOCConstraintDeclNode(LOCConstraintDeclNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by a <code>LTLConstraintDeclNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method
     * @return                 a list of strings that represents
     *                         <code> ltl name ( params ) ( formula ); </code>
     */
    public Object visitLTLConstraintDeclNode(LTLConstraintDeclNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by a <code>ELOCConstraintCallNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method
     * @return                 a list of strings that represents
     *                         <code> eloc name ( args ) </code>
     */

    public Object visitELOCConstraintCallNode(ELOCConstraintCallNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by a <code>LOCConstraintCallNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method
     * @return                 a list of strings that represents
     *                         <code> loc name ( args ) </code>
     */
    public Object visitLOCConstraintCallNode(LOCConstraintCallNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by a <code>LTLConstraintCallNode</code>.
     *
     * @param node        the node that is being visited
     * @param args        a list of arguments to this visit method
     * @return                 a list of strings that represents
     *                         <code> ltl name ( args ) </code>
     */
    public Object visitLTLConstraintCallNode(LTLConstraintCallNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by an <code>MinRateNode</code>.
     *
     * @param node      the node that is being visited
     * @param args      a list of arguments to this visit method
     * @return          a list of strings that represents
     *                  <code> minrate(quantity, event, value) </code>
     */
    public Object visitMinRateNode (MinRateNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by an <code>MaxRateNode</code>.
     *
     * @param node      the node that is being visited
     * @param args      a list of arguments to this visit method
     * @return          a list of strings that represents
     *                  <code> maxrate(quantity, event, value) </code>
     */
    public Object visitMaxRateNode (MaxRateNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by an <code>MinDeltaNode</code>.
     *
     * @param node      the node that is being visited
     * @param args      a list of arguments to this visit method
     * @return          a list of strings that represents
     *     <code> mindelta(quantity, event1, event2, value) </code>
     */
    public Object visitMinDeltaNode (MinDeltaNode node, LinkedList args) {
        return null;
    }

    /**
     * Returns the code represented by an <code>MaxDeltaNode</code>.
     *
     * @param node      the node that is being visited
     * @param args      a list of arguments to this visit method
     * @return          a list of strings that represents
     *     <code> maxdelta(quantity, event1, event2, value) </code>
     */
    public Object visitMaxDeltaNode (MaxDeltaNode node, LinkedList args) {
        return null;
    }

}
